/*
 * Copyright (C) 2022 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "oboe/OboeExtensions.h"
#include "aaudio/AAudioExtensions.h"

using namespace oboe;

bool OboeExtensions::isMMapSupported(){
    return AAudioExtensions::getInstance().isMMapSupported();
}

bool OboeExtensions::isMMapEnabled(){
    return AAudioExtensions::getInstance().isMMapEnabled();
}

int32_t OboeExtensions::setMMapEnabled(bool enabled){
    return AAudioExtensions::getInstance().setMMapEnabled(enabled);
}

bool OboeExtensions::isMMapUsed(oboe::AudioStream *oboeStream){
    return AAudioExtensions::getInstance().isMMapUsed(oboeStream);
}
