/*
  ==============================================================================

   This file is part of the JUCE library.
   Copyright (c) 2022 - Raw Material Software Limited

   JUCE is an open source library subject to commercial or open-source
   licensing.

   The code included in this file is provided under the terms of the ISC license
   http://www.isc.org/downloads/software-support-policy/isc-license. Permission
   To use, copy, modify, and/or distribute this software for any purpose with or
   without fee is hereby granted provided that the above copyright notice and
   this permission notice appear in all copies.

   JUCE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY, AND ALL WARRANTIES, WHETHER
   EXPRESSED OR IMPLIED, INCLUDING MERCHANTABILITY AND FITNESS FOR PURPOSE, ARE
   DISCLAIMED.

  ==============================================================================
*/

namespace juce::UnitTestCategories
{

    static const String analytics                  { "Analytics" };
    static const String audio                      { "Audio" };
    static const String audioProcessorParameters   { "AudioProcessorParameters" };
    static const String audioProcessors            { "AudioProcessors" };
    static const String blocks                     { "Blocks" };
    static const String compression                { "Compression" };
    static const String containers                 { "Containers" };
    static const String cryptography               { "Cryptography" };
    static const String dsp                        { "DSP" };
    static const String files                      { "Files" };
    static const String graphics                   { "Graphics" };
    static const String gui                        { "GUI" };
    static const String json                       { "JSON" };
    static const String maths                      { "Maths" };
    static const String memory                     { "Memory" };
    static const String midi                       { "MIDI" };
    static const String native                     { "Native" };
    static const String networking                 { "Networking" };
    static const String osc                        { "OSC" };
    static const String smoothedValues             { "SmoothedValues" };
    static const String streams                    { "Streams" };
    static const String text                       { "Text" };
    static const String threads                    { "Threads" };
    static const String time                       { "Time" };
    static const String values                     { "Values" };
    static const String xml                        { "XML" };

} // namespace juce::UnitTestCategories
